#!/usr/bin/perl -w

#
# tasuke_init.pl
# http://tasuke.dna.affrc.go.jp/
#
# Author: Jungsok Kim, Ryutaro Itoh
#
# Copyright (c) 2013 National Institute of Agrobiological Sciences   
#
# This software is released under the MIT License.
#
# http://opensource.org/licenses/mit-license.php
#

use strict;
use DBI;
use Getopt::Long;

my $RESTORE = '';
my $DB      = '';
my $HOST    = 'localhost';
my $USER    = '';
my $PASS    = '';
my $SAMPLE   = 'accession_ID';
my $ID = '';
my $RDBMS = '';
my $TABLESPACE = '';

################################################################
# Checking parameters and database connection
################################################################
GetOptions( 
            'db|database=s' => \$DB,
            'h|host=s'      => \$HOST,
            'u|user=s'      => \$USER,
            'b|backend=s'   => \$RDBMS,
            's|space=s'   => \$TABLESPACE,
            'p|password=s'  => \$PASS
);

if ( !$DB || !$USER ) {
    _error_method();
    exit;
}
if(!$RDBMS){
    $RDBMS = 'mysql';
}

my $connect;
if($RDBMS eq "oracle"){
    if(!$TABLESPACE){
        _error_method();
        exit;
    }
    $connect = DBI->connect('DBI:Oracle:host='.$HOST.';SERVICE_NAME='.$DB.';',$USER,$PASS)
    or die "Can't connect to database: $DBI::errstr\n";
    
    my $connect_ctxsys = DBI->connect('DBI:Oracle:host='.$HOST.';SERVICE_NAME='.$DB.';',"ctxsys","ctxsys");
    if(!$connect_ctxsys){
        print "\n----------------------------------------------------------------------\n";
        print "Unlocking a user named CTXSYS is required for text searching function.\n";
        print "Try following command after logged on database admin user.\n";
        print "> ALTER USER ctxsys IDENTIFIED ctxsys ACCOUNT UNLOCK\;\n";
        print "----------------------------------------------------------------------\n";
        _exit(1);
    }else{
        $connect_ctxsys->disconnect or warn "Disconnection error: $DBI::errstr\n";
    }
    if(!_check_exists_tablespace($TABLESPACE)){
        print "Tablespace: '$TABLESPACE' is not found.";
        _exit(1);
    }
}else{
    $connect= DBI->connect('DBI:mysql:database='.$DB.';host='.$HOST,$USER,$PASS)
    or die "Can't connect to database: $DBI::errstr\n";
}

###############################################################
# Checking default tables
# [1] sys_track
# [2] sys_snpEff
# [3] sys_snpblock
###############################################################
_check_default_tables();

###############################################################
# Installing Started 
###############################################################
### Starting chromosome & accession list tables
$SIG{'INT'} = '_SIGINThandler';
print "\n\n  ------------------------------------------------";
print "\n   Installing TASUKE multiple genome browser started";
print "\n  ------------------------------------------------";
print "\n\n  Input csv file about chromosome's information below.";
if ( ! defined _check_exists_table("sys_chromosome") ) {
    _first_chrs();
} else {
    print "  Completed.";
}
if ( ! defined _check_exists_table("sys_accession") ) {
    if($RDBMS eq "oracle"){
        _dbexe("CREATE TABLE sys_accession (
        accession_id varchar2(40) NOT NULL,
        accession_db varchar2(6) DEFAULT NULL,
        accession_name varchar2(40) DEFAULT NULL,
        variety varchar2(40) DEFAULT NULL,
        sub_variety varchar2(40) DEFAULT NULL,
        origin varchar2(30) DEFAULT NULL,
        origin_2 varchar2(30) DEFAULT NULL,
        type varchar2(40) DEFAULT NULL,
        num_reads number(16,0) DEFAULT 0,
        PRIMARY KEY (accession_id)
        )TABLESPACE $TABLESPACE");
        _dbexe("CREATE OR REPLACE PUBLIC SYNONYM sys_accession FOR $USER.sys_accession");
    }else{
        _dbexe("DROP TABLE IF EXISTS  sys_accession");
        _dbexe("CREATE TABLE `sys_accession` (
            `accession_id` varchar(40) NOT NULL,
            `accession_db` varchar(6) DEFAULT NULL,
            `accession_name` varchar(40) DEFAULT NULL,
            `variety` varchar(40) DEFAULT NULL,
            `sub_variety` varchar(40) DEFAULT NULL,
            `origin` varchar(20) DEFAULT NULL,
            `origin_2` varchar(30) DEFAULT NULL,
            `type` varchar(40) DEFAULT NULL,
            `num_reads` bigint DEFAULT 0,
             PRIMARY KEY (`accession_id`)
            )");
    }
}

print "\n\n  -----------\n";
print "   Next step\n";
print "  -----------\n\n";
print "   1. Input Accession information\n";
print "     # tasuke_accession.pl -c -db $DB -u $USER -p $PASS\n\n";
print "   2. Input Reference fasta\n";
print "     # tasuke_ref.pl -f [file(.fa)] -db $DB -u $USER -p $PASS\n\n";
print "   3. Input VARIANT file as count of accessions you inputted.\n";
print "     # tasuke_variant_vcf.pl -n [$SAMPLE] -f [file(.vcf)] -db $DB -u $USER -p $PASS\n\n";
print "   4. Input DEPTH file as count of accessions you inputted. (Optional)\n";
print "     # tasuke_tsv_db.pl -n [$SAMPLE] -f [file(.tsv)] -db $DB -u $USER -p $PASS\n\n";
print "   5. Input tsv file for the custom track as count of accessions you inputted. (Optional)\n";
print "     # tasuke_tsv_db.pl -c -n [$SAMPLE] -f [file(.tsv)] -db $DB -u $USER -p $PASS\n\n";
print "   6. Input gene tracks (Optional)\n";
print "     # tasuke_track.pl -t [track_name] -f [file(.gff)] -db $DB -u $USER -p $PASS\n\n";
print "\n#######################################################################\n";

_exit();

###############################################################
# Set csv file for chromosomes
###############################################################
sub _first_chrs {
    my $F;
    print "\n\n    +----  CSV  -----------------------+";
    print "\n    | chr01,43270923,16610866,17243770 |";
    print "\n    | chr02,35937250,13872411,13541821 |";
    print "\n      ..";
    print "\n    | chr12,27531856,11761737,12103486 |";
    print "\n    +----------------------------------+";
    print "\n    [Formats] chrs, size of chrs, start of centromere, end of centromere\n\n";

    ## Require csv file
    print "> Where is the csv file? # ";
    while (defined(my $line = <STDIN>)){
        chomp($line);
        $line=~s/^\s*(.*?)\s*$/$1/;
        $F=$line;
        if ( $F ne "" && -e $F ) {
            if ( _check_chrs_csv($F) ) {
                print "It is wrong format.\n";
            } elsif( _check_sequence_name ($F) ){
                print "TASUKE does not accept the chromosome name which is included \".\"(dot).\n";
            } else {
                last;
            }
        } elsif ( $F ne "" && ! -e $F ) {
            print "File does not exists.\n";
        }
        print "> Where is the csv file? # ";
    }

    ## Confirmation
    if(_confirm_chrs($F) !=0){_exit(1);}
    print "> Are you sure creating database?[y|n] ";
    while (defined(my $line = <STDIN>)){
        chomp($line);
        if ( $line eq "y" || $line eq "Y" ) {
            last;
        } elsif ( $line eq "n" || $line eq "N" ) {
            print "finished.\n";
            exit;
        } else {
            print "> Are you sure creating database?[y|n] ";
        }
    }
    _create_chrs_table($F);
}

sub _exit {
    my ($arg) = @_;
    my $code = 0;
    if(defined $arg){$code = $arg;}
    $connect->disconnect or warn "Disconnection error: $DBI::errstr\n";
    exit($code);
}


###############################################################
# Recognization csv file
###############################################################
sub _confirm_chrs {
    my ($f) = @_;
    my $e = 0;
    my $num = 0;
    print "\n   * Have you check the list you allocate.[".$f."]\n";
    print "   +----------------------------------------------------------------+\n";
    print "   | chrs  | size of chrs | start of centromere | end of centromere |\n";
    print "   +----------------------------------------------------------------+\n";
    open IN, '<', $f or die "file open error: $!";
    while (<IN>) {
        chomp($_);
        my @line=split(/\,/, $_);
        printf("   |%6s |%13d |%20s |%18s |\n",$line[0],$line[1],$line[2],$line[3]);
        $num++;
    }
    print "   +----------------------------------------------------------------+\n";
    if($num > 49){
        $e = 1;
        print "\n   * WARNING\n";
        print "     Over 50 sequences were detected in this file.\n";
        print "     TASUKE can accept under 50 sequences.\n";
        print "Exit\n";
    }else{
        print "\n   * WARNING\n";
        print "     Don't change column at [chrs], [size of chrs] on the database after any of this process.\n";
        print "     If you found miss some values, select [n] and again this process now.\n";
        print "     \n";
    }
    close IN;
    return $e;
}

###############################################################
# Checking chromosome file
###############################################################
sub _check_chrs_csv {
    my ($f)= @_;
    my $e=0;
    open IN, '<', $f or die "file open error: $!";
    while (<IN>) {
        chomp($_);
        my @line=split(/,/, $_);
        $e=1 if $#line != 3; 
    }
    close IN;
    return $e;
}

###############################################################
# Checking chromosome file
###############################################################
sub _check_sequence_name {
    my ($f)= @_;
    my $e=0;
    open IN, '<', $f or die "file open error: $!";
    while (<IN>) {
        chomp($_);
        my @line=split(/,/, $_);
        if($line[0] =~ /\./){$e=1;}
    }
    close IN;
    return $e;
}

###############################################################
# Creating chromosome table
###############################################################
sub _create_chrs_table {
    my ($f)= @_;
    my $e=0;
    
    if($RDBMS eq "oracle"){
        if ( defined _check_exists_table("sys_chromosome") ) {
            _dbexe("DROP TABLE sys_chromosome cascade constraint");
        }
        _dbexe("CREATE TABLE sys_chromosome (
                CONSTRAINT KEY_SYS_CHR seq number(6,0) PRIMARY KEY,
                id VARCHAR2(50) NOT NULL,
                seq_size number(11,0) NOT NULL,
                centromere_s number(11,0),  
                centromere_e number(11,0)
                )TABLESPACE $TABLESPACE");
        if(! _check_exists_seq("SEQ_SYS_CHROMOSOME")){
            _dbexe("CREATE SEQUENCE SEQ_SYS_CHROMOSOME INCREMENT BY 1 START WITH 1");
        }
        _dbexeIgnNamePh("CREATE OR REPLACE TRIGGER TRG_SYS_CHROMOSOME 
                        BEFORE INSERT ON sys_chromosome FOR EACH ROW 
                        BEGIN SELECT SEQ_SYS_CHROMOSOME.NEXTVAL INTO :NEW.seq FROM DUAL; 
                        END; 
                        ");
        _dbexe("CREATE OR REPLACE PUBLIC SYNONYM sys_chromosome FOR $USER.sys_chromosome");
        open IN, '<', $f or die "file open error: $!";
        while (<IN>) {
            chomp($_);
            my @line=split(/,/, $_);
            $line[0] =~ s/^\s*(.*?)\s*$/$1/;
            $line[1] =~ s/^\s*(.*?)\s*$/$1/;
            $line[2] =~ s/^\s*(.*?)\s*$/$1/;
            $line[3] =~ s/^\s*(.*?)\s*$/$1/;
            _dbexe("INSERT INTO sys_chromosome(id,seq_size,centromere_s,centromere_e)
                    VALUES('".$line[0]."',".$line[1].",".$line[2].",".$line[3].")");
        }
        close IN;

        ## Creating table sys_fasta
        my @c=_chrs();
        if ( ! defined _check_exists_table("sys_fasta") ) {
            my $s="CREATE TABLE sys_fasta ( pos number(11,0) PRIMARY KEY,";
            for ( my $i=0; $i<=$#c; $i++) {
                $s = $s . $c[$i] . "_f VARCHAR2(1)";
                $s = $s . "," if ( $i < $#c);
            }
            $s=$s.")TABLESPACE $TABLESPACE";
            _dbexe($s);
            #_dbexe("CREATE SEQUENCE SEQ_SYS_FASTA INCREMENT BY 1 START WITH 1");
            #_dbexeIgnNamePh("CREATE OR REPLACE TRIGGER TRG_SYS_FASTA
            #                BEFORE INSERT ON sys_fasta FOR EACH ROW 
            #                BEGIN SELECT SEQ_SYS_FASTA.NEXTVAL INTO :NEW.pos FROM DUAL; 
            #                END; 
            #                ");
            _dbexe("CREATE OR REPLACE PUBLIC SYNONYM sys_fasta FOR $USER.sys_fasta");
        }
        print "OK. ".int($#c+1)." chromosomes is setted.";
    }else{
        _dbexe("DROP TABLE iF EXISTS  sys_chromosome");
        _dbexe("CREATE TABLE sys_chromosome (
                seq SMALLINT AUTO_INCREMENT PRIMARY KEY,
                id VARCHAR(50) NOT NULL,
                size INT NOT NULL,
                centromere_s INT,  
                centromere_e INT
            )");
        open IN, '<', $f or die "file open error: $!";
        while (<IN>) {
            chomp($_);
            my @line=split(/,/, $_);
            $line[0] =~ s/^\s*(.*?)\s*$/$1/;
            $line[1] =~ s/^\s*(.*?)\s*$/$1/;
            $line[2] =~ s/^\s*(.*?)\s*$/$1/;
            $line[3] =~ s/^\s*(.*?)\s*$/$1/;
            _dbexe("INSERT INTO sys_chromosome(id,size,centromere_s,centromere_e)
                     VALUES('".$line[0]."',".$line[1].",".$line[2].",".$line[3].")");
        }
        close IN;

        ## Creating table sys_fasta
        my @c=_chrs();
        if ( ! defined _check_exists_table("sys_fasta") ) {
            my $s="CREATE TABLE sys_fasta ( pos INT PRIMARY KEY AUTO_INCREMENT,";
            for ( my $i=0; $i<=$#c; $i++) {
                $s = $s . $c[$i] . "_f CHAR(1)";
                $s = $s . "," if ( $i < $#c);
            }
            $s=$s.")";
            _dbexe($s);
        }
        print "OK. ".int($#c+1)." chromosomes is setted.";
    }
    return $e;
}

### Set chromosome ################################################
sub _chrs {
    my $val="";
    my @c;
    my $query = "SELECT id FROM sys_chromosome";
    my $query_handle = $connect->prepare($query);
    $query_handle->execute();
    $query_handle->bind_columns(undef, \$val);
    while($query_handle->fetch()) {
        push @c, $val;
    }
    return @c;
}

### Checking default(base) tables  #########################################
sub _check_default_tables {
    if($RDBMS eq "oracle"){
        my $val1 = $connect->selectrow_array("select table_name from user_tables where tablespace_name = '$TABLESPACE' and table_name = 'SYS_SNPEFF'");
        if ( ! defined $val1 ) {
            _dbexe("CREATE TABLE sys_snpEff(
                    eff_id number(6,0) NOT NULL,
                    eff_name varchar2(100) DEFAULT NULL,
                    eff_grp number(6,0) DEFAULT NULL,
                    CONSTRAINT KEY_SYS_SNPEFF PRIMARY KEY (eff_id)
                    ) TABLESPACE $TABLESPACE");
            if(! _check_exists_seq("SEQ_SYS_SNPEFF")){
                _dbexe("CREATE SEQUENCE SEQ_SYS_SNPEFF INCREMENT BY 1 START WITH 1");
            }
            _dbexeIgnNamePh("CREATE OR REPLACE TRIGGER TRG_SYS_SNPEFF 
                            BEFORE INSERT ON sys_snpEff FOR EACH ROW 
                            BEGIN SELECT SEQ_SYS_SNPEFF.NEXTVAL INTO :NEW.eff_id FROM DUAL; 
                            END; 
                            ");
            _dbexe("INSERT ALL 
                    INTO sys_snpEff VALUES(1,'INTERGENIC',0)
                    INTO sys_snpEff VALUES(2,'UPSTREAM',1)
                    INTO sys_snpEff VALUES(3,'DOWNSTREAM',1)
                    INTO sys_snpEff VALUES(4,'UTR_5_PRIME',2)
                    INTO sys_snpEff VALUES(5,'UTR_3_PRIME',2)
                    INTO sys_snpEff VALUES(6,'REGULATION',2)
                    INTO sys_snpEff VALUES(7,'INTRAGENIC',2)
                    INTO sys_snpEff VALUES(8,'GENE',2)
                    INTO sys_snpEff VALUES(9,'INTRON_CONSERVED',2)
                    INTO sys_snpEff VALUES(10,'INTRON',2)
                    INTO sys_snpEff VALUES(11,'INTERGENIC_CONSERVED',2)
                    INTO sys_snpEff VALUES(12,'SYNONYMOUS_START',3)
                    INTO sys_snpEff VALUES(13,'NON_SYNONYMOUS_START',3)
                    INTO sys_snpEff VALUES(14,'START_GAINED',3)
                    INTO sys_snpEff VALUES(15,'SYNONYMOUS_CODING',3)
                    INTO sys_snpEff VALUES(16,'SYNONYMOUS_STOP',3)
                    INTO sys_snpEff VALUES(17,'EXON',3)
                    INTO sys_snpEff VALUES(18,'CDS',3)
                    INTO sys_snpEff VALUES(19,'TRANSCRIPT',3)
                    INTO sys_snpEff VALUES(20,'NON_SYNONYMOUS_CODING',4)
                    INTO sys_snpEff VALUES(21,'CODON_CHANGE',4)
                    INTO sys_snpEff VALUES(22,'CODON_INSERTION',4)
                    INTO sys_snpEff VALUES(23,'CODON_CHANGE_PLUS_CODON_INSERTION',4)
                    INTO sys_snpEff VALUES(24,'CODON_DELETION',4)
                    INTO sys_snpEff VALUES(25,'CODON_CHANGE_PLUS_CODON_DELETION',4)
                    INTO sys_snpEff VALUES(26,'UTR_5_DELETED',4)
                    INTO sys_snpEff VALUES(27,'UTR_3_DELETED',4)
                    INTO sys_snpEff VALUES(28,'SPLICE_SITE_ACCEPTOR',5)
                    INTO sys_snpEff VALUES(29,'SPLICE_SITE_DONOR',5)
                    INTO sys_snpEff VALUES(30,'START_LOST',5)
                    INTO sys_snpEff VALUES(31,'EXON_DELETED',5)
                    INTO sys_snpEff VALUES(32,'FRAME_SHIFT',5)
                    INTO sys_snpEff VALUES(33,'STOP_GAINED',5)
                    INTO sys_snpEff VALUES(34,'STOP_LOST',5)
                    INTO sys_snpEff VALUES(35,'RARE_AMINO_ACID',5) 
                    SELECT * from dual");
            _dbexe("CREATE OR REPLACE PUBLIC SYNONYM sys_snpEff FOR $USER.sys_snpEff");
        }
        my $val2 = $connect->selectrow_array("select table_name from user_tables where tablespace_name = '$TABLESPACE' and table_name = 'SYS_TRACK'");
        if ( ! defined $val2 ) {
            _dbexe("CREATE TABLE sys_track (
                   seq number(38,0) NOT NULL ,
                   track varchar2(50) DEFAULT NULL,
                   chr varchar2(50) DEFAULT NULL,
                   id varchar2(50) DEFAULT NULL,
                   name varchar2(59) DEFAULT NULL,
                   parent number(1,0) DEFAULT NULL,
                   type varchar2(40) DEFAULT NULL,
                   pos number(11,0) DEFAULT NULL,
                   end number(11,0) DEFAULT NULL,
                   strand number(4,0) DEFAULT NULL,
                   note varchar2(4000) DEFAULT NULL,
                   others varchar2(4000) DEFAULT NULL,
                   CONSTRAINT KEY_SYS_TRACK PRIMARY KEY (seq)
                    )TABLESPACE $TABLESPACE");
            if(! _check_exists_seq("SEQ_SYS_TRACK")){
                _dbexe("CREATE SEQUENCE SEQ_SYS_TRACK INCREMENT BY 1 START WITH 1");
            }
            _dbexeIgnNamePh("CREATE OR REPLACE TRIGGER TRG_SYS_TRACK 
                            BEFORE INSERT ON sys_track FOR EACH ROW 
                            BEGIN SELECT SEQ_SYS_TRACK.NEXTVAL INTO :NEW.seq FROM DUAL; 
                            END; 
                            ");
            _dbexe("CREATE INDEX sys_track_idx on sys_track(id) TABLESPACE $TABLESPACE");
            _dbexe("CREATE INDEX sys_track_pidx on sys_track(chr, pos) TABLESPACE $TABLESPACE");
            if(! _check_exists_pref("GFF_NOTE_LEXER") && ! _check_exists_pref("GFF_OTHERS_LEXER") && ! _check_exists_pref("GFF_ID_LEXER")){
                _dbexeIgnNamePh("BEGIN
                                ctx_ddl.create_preference('GFF_ID_LEXER','BASIC_LEXER');
                                ctx_ddl.create_preference('GFF_NOTE_LEXER','BASIC_LEXER');
                                ctx_ddl.create_preference('GFF_OTHERS_LEXER','BASIC_LEXER');
                                END;
                                ");
            }
            _dbexe("CREATE INDEX sys_track_id_idx on sys_track(id) indextype is ctxsys.context parameters('LEXER GFF_ID_LEXER')");
            _dbexe("CREATE INDEX sys_track_note_idx on sys_track(note) indextype is ctxsys.context parameters('LEXER GFF_NOTE_LEXER')");
            _dbexe("CREATE INDEX sys_track_others_idx on sys_track(others) indextype is ctxsys.context parameters('LEXER GFF_OTHERS_LEXER')");
            _dbexe("CREATE OR REPLACE PUBLIC SYNONYM sys_track FOR $USER.sys_track");
        }
        my $val3 = $connect->selectrow_array("select table_name from user_tables where tablespace_name = '$TABLESPACE' and table_name = 'SYS_SNPBLOCK'");
        if ( ! defined $val3 ) {
            _dbexe("CREATE TABLE sys_snpblock (
                  accession varchar2(40) DEFAULT NULL,
                  rate float DEFAULT NULL,
                  peak varchar2(20) DEFAULT NULL)TABLESPACE $TABLESPACE");
            _dbexe("CREATE OR REPLACE PUBLIC SYNONYM sys_snpblock FOR $USER.sys_snpblock");
        }
    }else{
        my $val1 = $connect->selectrow_array("SHOW TABLES LIKE 'sys_snpEff'");
        if ( ! defined $val1 ) {
        _dbexe("CREATE TABLE `sys_snpEff` (
                `eff_id` smallint(6) NOT NULL AUTO_INCREMENT,
                `eff_name` varchar(100) DEFAULT NULL,
                `eff_grp` smallint(6) DEFAULT NULL,
                PRIMARY KEY (`eff_id`))");
        _dbexe("INSERT INTO `sys_snpEff` VALUES
                (1,'INTERGENIC',0),(2,'UPSTREAM',1),(3,'DOWNSTREAM',1),(4,'UTR_5_PRIME',2),
                (5,'UTR_3_PRIME',2),(6,'REGULATION',2),(7,'INTRAGENIC',2),(8,'GENE',2),
                (9,'INTRON_CONSERVED',2),(10,'INTRON',2),(11,'INTERGENIC_CONSERVED',2),
                (12,'SYNONYMOUS_START',3),(13,'NON_SYNONYMOUS_START',3),(14,'START_GAINED',3),
                (15,'SYNONYMOUS_CODING',3),(16,'SYNONYMOUS_STOP',3),(17,'EXON',3),(18,'CDS',3),
                (19,'TRANSCRIPT',3),(20,'NON_SYNONYMOUS_CODING',4),(21,'CODON_CHANGE',4),
                (22,'CODON_INSERTION',4),(23,'CODON_CHANGE_PLUS_CODON_INSERTION',4),(24,'CODON_DELETION',4),
                (25,'CODON_CHANGE_PLUS_CODON_DELETION',4),(26,'UTR_5_DELETED',4),(27,'UTR_3_DELETED',4),
                (28,'SPLICE_SITE_ACCEPTOR',5),(29,'SPLICE_SITE_DONOR',5),(30,'START_LOST',5),(31,'EXON_DELETED',5),
                (32,'FRAME_SHIFT',5),(33,'STOP_GAINED',5),(34,'STOP_LOST',5),(35,'RARE_AMINO_ACID',5)");
        }

        my $val2 = $connect->selectrow_array("show tables like 'sys_track'");
        if ( ! defined $val2 ) {
        _dbexe("CREATE TABLE `sys_track` (
               `seq` int(11) NOT NULL AUTO_INCREMENT,
               `track` varchar(50) DEFAULT NULL,
               `chr` varchar(50) DEFAULT NULL,
               `id` varchar(50) DEFAULT NULL,
               `name` varchar(59) DEFAULT NULL,
               `parent` int(1) DEFAULT NULL,
               `type` varchar(40) DEFAULT NULL,
               `pos` int(11) DEFAULT NULL,
               `end` int(11) DEFAULT NULL,
               `strand` tinyint(4) DEFAULT NULL,
               `note` text ,
               `others` text ,
               PRIMARY KEY (`seq`),
               KEY `idx` (`id`),
               KEY `p_idx` (`chr`,`pos`)
            )ENGINE MyISAM");
        }

        my $val3 = $connect->selectrow_array("show tables like 'sys_snpblock'");
        if ( ! defined $val3 ) {
        _dbexe("CREATE TABLE `sys_snpblock` (
                  `accession` varchar(40) DEFAULT NULL,
                  `rate` float DEFAULT NULL,
                  `peak` varchar(20) DEFAULT NULL)");
        }
    }
}

### Checking list table  #########################################
sub _get_list{
    my $val="";
    my @id;
    my $query = "SELECT accession_id FROM sys_accession";
    my $query_handle = $connect->prepare($query);
    $query_handle->execute();
    $query_handle->bind_columns(undef, \$val);
    while($query_handle->fetch()) {
        push @id, $val;
    }
    return \@id;
}

### Checking chromosome table ####################################
sub _check_exists_table {
    my ($t) = @_;
    my $val;
    if($RDBMS eq "oracle"){
        $t = uc($t);
        $val = $connect->selectrow_array("select table_name from user_tables where tablespace_name = '$TABLESPACE' and table_name = '$t'");
    }else{
        $val = $connect->selectrow_array("show tables like '$t'");
    }
    return $val;
}

### Checking oracel sequence ####################################
sub _check_exists_seq {
    my ($t) = @_;
    $t = uc($t);
    my $val = $connect->selectrow_array("select sequence_name from user_sequences where sequence_name = '$t'");
    return $val;
}

### Checking preference ####################################
sub _check_exists_pref {
    my ($t) = @_;
    $t = uc($t);
    my $val = $connect->selectrow_array("select pre_name from ctx_user_preferences where pre_name = '$t'");
    return $val;
}

### Checking tablespace ####################################
sub _check_exists_tablespace {
    my ($t) = @_;
    $t = uc($t);
    my $val = $connect->selectrow_array("select tablespace_name from user_tablespaces where tablespace_name = '$t'");
    return $val;
}

### Execute DB ###################################################
sub _dbexe {
    my ($query) = @_;
    my $query_handle = $connect->prepare($query);
    $query_handle->execute() or warn "Excute error: $DBI::errstr\n";
}

sub _dbexeIgnNamePh{
    my ($query) = @_;
    my $query_handle = $connect;
    $query_handle->{odbc_ignore_named_placeholders} = 1;
    $query_handle->do($query) or warn "Excute error: $DBI::errstr\n";#odbc_ignore_named_placeholders
}

#### Time #######################################
sub _print_time {
    my ($str) = @_;
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime;
    $year += 1900;
    $mon += 1;
    my $datetime = sprintf "%04d-%02d-%02d %02d:%02d:%02d", $year, $mon, $mday, $hour, $min, $sec;
    print "$str : $datetime\n";
}

#### Usage #######################################
sub _error_method {
    print "\nFor MySQL database:\n";
    print "tasuke_init.pl -db [database] -u [user] -p [password]\n\n";
    print "For Oracle database:\n";
    print "tasuke_init.pl -db [sid or service_name] -b oracle -s [table_space_name] -u [user] -p [password]\n\n";
}

### SIGINT ###
sub _SIGINThandler{
    print "\n\nFinished.\n";
    _exit(1);
}